/*
 This file is part of conc.

    Conc is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Conc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with conc; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gconf/gconf.h>
#include <gnome.h>
#include "conc.h"
#include "conc-ui.h"
#include "tuple.h"

enum conc_configure_server_new_widget_index {
  CONFIG_SERVER_NEW_local,
  CONFIG_SERVER_NEW_name,
  CONFIG_SERVER_NEW_termserv,
  CONFIG_SERVER_NEW_button,
  CONFIG_SERVER_NEW_dialog,
  CONFIG_SERVER_NEW_END
};

enum conc_configure_console_widget_index {
  CONFIG_CONSOLE_name,
  CONFIG_CONSOLE_server,
  CONFIG_CONSOLE_device,
  CONFIG_CONSOLE_bits8,
  CONFIG_CONSOLE_bits7,
  CONFIG_CONSOLE_stopbits1,
  CONFIG_CONSOLE_stopbits2,
  CONFIG_CONSOLE_none,
  CONFIG_CONSOLE_even,
  CONFIG_CONSOLE_odd,
  CONFIG_CONSOLE_transmit,
  CONFIG_CONSOLE_receive,
  CONFIG_CONSOLE_log,
  CONFIG_CONSOLE_button,
  CONFIG_CONSOLE_dialog,
  CONFIG_CONSOLE_spec,
  CONFIG_CONSOLE_END
};

enum conc_configure_widget_index {
  CONFIG_server_list,
  CONFIG_server_remove,
  CONFIG_console_list,
  CONFIG_console_edit,
  CONFIG_console_remove,
  CONFIG_END
};

void
conc_configure_list_server_changed(void)
{
  gchar *sp[1];
  gint i,row;
  GtkWidget **a=G->config_app_w;

  gtk_clist_clear(GTK_CLIST(a[CONFIG_server_list]));
  gtk_widget_set_sensitive(a[CONFIG_server_remove],FALSE);
  for (i=0;i<conc_list_length(G->server_list);i++) {
    sp[0]=g_strdup(conc_list(G->server_list,i));
    row=gtk_clist_append(GTK_CLIST(a[CONFIG_server_list]),sp);
  }
}

void
conc_configure_list_console_changed(void)
{
  gchar *sp[1];
  gint i,row;
  GtkWidget **a=G->config_app_w;
  gchar *s;

  gtk_clist_clear(GTK_CLIST(a[CONFIG_console_list]));
  gtk_widget_set_sensitive(a[CONFIG_console_remove],FALSE);
  gtk_widget_set_sensitive(a[CONFIG_console_edit],FALSE);
  for (i=0;i<conc_list_length(G->console_list);i++) {
    s=(gchar *)conc_list(G->console_list,i);
    sp[0]=&s[2];
    row=gtk_clist_append(GTK_CLIST(a[CONFIG_console_list]),sp);
  }
}

static void
conc_configure_remove_server_done(ConcChannel *chan, gchar *line,
                                  gpointer cd, gpointer d)
{
  GtkWidget *w;
  gchar *s,*t;

  if (line[0]=='o') {
    g_free(line);
    return;
  }
  t=g_strdelimit(&line[1],"\r",'\n');
  s=g_strdup_printf("Error removing server:\n%s",t);
  g_free(line);
  w=gnome_message_box_new(s,GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
  g_free(s);
  gtk_widget_show_all(w);
  gnome_dialog_run_and_close(GNOME_DIALOG(w));
}

static void
conc_configure_remove_server(GtkButton *button,gpointer data)
{
  GtkWidget **a;
  gchar *s,*t;

  a=G->config_app_w;
  if (!gtk_clist_get_text(GTK_CLIST(a[CONFIG_server_list]),
                          G->config_app_server_selection,0,&t)) {
    g_warning("could not fetch data from clist");
    return;
  }
  s=g_strdup_printf("cds%s",t);
  conc_channel_command(G->srv_chan,s,conc_configure_remove_server_done,NULL);
  g_free(s);
}

static int
conc_configure_new_server_close(GnomeDialog *dialog, gpointer data)
{
  GtkWidget **a=(GtkWidget **)data;

  gtk_widget_set_sensitive(a[CONFIG_SERVER_NEW_button],TRUE);
  g_free(a);
  return FALSE;
}

void
conc_configure_new_server_cancel_button(GtkButton *b, gpointer data)
{
  GtkWidget **a=(GtkWidget **)data;

  gnome_dialog_close(GNOME_DIALOG(a[CONFIG_SERVER_NEW_dialog]));
}

void
conc_configure_new_server_done(ConcChannel *chan, gchar *line,
                               gpointer cd, gpointer d)
{
  GtkWidget *w;
  gchar *s;

  if (line[0]=='e') {
    s=g_strdup_printf("Error opening server:\n%s",&line[1]);
    w=gnome_message_box_new(s,GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
    g_free(s);
    gtk_widget_show_all(w);
    gnome_dialog_run_and_close(GNOME_DIALOG(w));
  }
  g_free(line);
}

void
conc_configure_new_server_ok_button(GtkButton *b, gpointer data)
{
  GtkWidget **a=(GtkWidget **)data;
  GtkWidget *w;
  gchar *name, *termserv;
  gchar *s;

  if GTK_WIDGET_STATE(a[CONFIG_SERVER_NEW_local]) {
    s=g_strdup("cnsl");
  } else {
    name=gtk_editable_get_chars(GTK_EDITABLE(a[CONFIG_SERVER_NEW_name]),0,-1);
    if (!strlen(name)) {
      g_free(name);
      w=gnome_message_box_new(_("Server name required."),GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,GNOME_STOCK_BUTTON_CANCEL,NULL);
      gtk_widget_show_all(w);
      if (gnome_dialog_run_and_close(GNOME_DIALOG(w))==1) gnome_dialog_close(GNOME_DIALOG(a[CONFIG_SERVER_NEW_dialog]));
      return;
    }
    termserv=gtk_editable_get_chars(GTK_EDITABLE(a[CONFIG_SERVER_NEW_termserv]),0,-1);
    if (!strlen(termserv)) {
      g_free(name);
      g_free(termserv);
      w=gnome_message_box_new(_("Command required."),GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,GNOME_STOCK_BUTTON_CANCEL,NULL);
      gtk_widget_show_all(w);
      if (gnome_dialog_run_and_close(GNOME_DIALOG(w))==1) gnome_dialog_close(GNOME_DIALOG(a[CONFIG_SERVER_NEW_dialog]));
      return;
    }
    s=g_strdup_printf("cnsr%s\r%s",name,termserv);
    g_free(name); g_free(termserv);
  }
  conc_channel_command(G->srv_chan,s,conc_configure_new_server_done,NULL);
  g_free(s);
  gnome_dialog_close(GNOME_DIALOG(a[CONFIG_SERVER_NEW_dialog]));
}

void
conc_configure_local_toggled(GtkToggleButton *button, gpointer data)
{
  GtkWidget **a=(GtkWidget **)data;
  gboolean active;

  if (GTK_WIDGET_STATE(GTK_WIDGET(button))==GTK_STATE_ACTIVE) active=FALSE; else active=TRUE;
  gtk_widget_set_sensitive(a[CONFIG_SERVER_NEW_name],active);
}

static void
conc_configure_new_server(GtkButton *button,gpointer data)
{
  GtkWidget *w,*x;
  GtkWidget **a=g_new0(GtkWidget *,CONFIG_SERVER_NEW_END);

  a[CONFIG_SERVER_NEW_button]=GTK_WIDGET(button);
  gtk_widget_set_sensitive((GtkWidget *)button,FALSE);
  a[CONFIG_SERVER_NEW_dialog]=gnome_dialog_new(_("New Server"),GNOME_STOCK_BUTTON_OK,GNOME_STOCK_BUTTON_CANCEL,NULL);
  gnome_dialog_set_parent(GNOME_DIALOG(a[CONFIG_SERVER_NEW_dialog]),GTK_WINDOW(G->app));
  gtk_signal_connect(GTK_OBJECT(a[CONFIG_SERVER_NEW_dialog]), "close",
                     GTK_SIGNAL_FUNC(conc_configure_new_server_close),
                     a);
  gnome_dialog_button_connect(GNOME_DIALOG(a[CONFIG_SERVER_NEW_dialog]),0,conc_configure_new_server_ok_button,a);
  gnome_dialog_button_connect(GNOME_DIALOG(a[CONFIG_SERVER_NEW_dialog]),1,conc_configure_new_server_cancel_button,a);

  a[CONFIG_SERVER_NEW_local]=gtk_check_button_new_with_label(_("Server is local to master"));
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(a[CONFIG_SERVER_NEW_dialog])->vbox),a[CONFIG_SERVER_NEW_local],TRUE,TRUE,GNOME_PAD_SMALL);

  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(a[CONFIG_SERVER_NEW_dialog])->vbox),x,TRUE,TRUE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Server name:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  a[CONFIG_SERVER_NEW_name]=gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(x),a[CONFIG_SERVER_NEW_name],TRUE,TRUE,0);

  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(a[CONFIG_SERVER_NEW_dialog])->vbox),x,TRUE,TRUE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Command:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  a[CONFIG_SERVER_NEW_termserv]=gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(x),a[CONFIG_SERVER_NEW_termserv],TRUE,TRUE,0);

  gtk_signal_connect(GTK_OBJECT(a[CONFIG_SERVER_NEW_local]),"toggled",GTK_SIGNAL_FUNC(conc_configure_local_toggled),a);
  gtk_widget_show_all(a[CONFIG_SERVER_NEW_dialog]);
}

static GtkWidget
*speed_menu(void)
{
  GtkWidget *m;
  gchar *speeds[] = { "0", "50", "75", "110", "134", "150", "200", "300",
                     "600", "1200", "1800", "2400", "4800", "9600", "19200",
                     "38400", "57600", "115200", "230400", NULL
  };
  gint i;

  m=gtk_menu_new();
  for (i=0;speeds[i];i++) {
    gtk_menu_append(GTK_MENU(m),gtk_menu_item_new_with_label(speeds[i]));
  }
  return m;
}

static gint
get_speed(guint i)
{
  gint speeds[] = { 0, 50, 75, 110, 134, 150, 200, 300,
                    600, 1200, 1800, 2400, 4800, 9600, 19200,
                    38400, 57600, 115200, 230400
  };

  return speeds[i];
}

static gint
set_speed(guint speed)
{
  gint speeds[] = { 0, 50, 75, 110, 134, 150, 200, 300,
                    600, 1200, 1800, 2400, 4800, 9600, 19200,
                    38400, 57600, 115200, 230400, -1
  };
  gint i;

  for (i=0;speeds[i]>-1;i++) if (speeds[i]==speed) break;
  return speeds[i]>-1?i:-1;
}

static void
conc_configure_console_show_greylog(GtkButton *button, gpointer data)
{
  GPtrArray *t=data;
  GtkWidget *log=GTK_WIDGET(conc_tuple_get(t,0));
  GtkWidget *name=GTK_WIDGET(conc_tuple_get(t,1));
  gchar *b=conc_tuple_get(t,2);
  gchar *s;
  gint i;

  gtk_widget_set_sensitive(GTK_WIDGET(log),FALSE);
  gtk_editable_delete_text(GTK_EDITABLE(log),0,-1);
  if (b[0]=='d') {
    s=gtk_editable_get_chars(GTK_EDITABLE(name),0,-1);
    i=0;
    gtk_editable_insert_text(GTK_EDITABLE(log),s,strlen(s),&i);
    g_free(s);
  }
}

static void
conc_configure_console_show_log(GtkButton *button, gpointer data)
{
  gtk_widget_set_sensitive(GTK_WIDGET(data),TRUE);
}

static void
conc_configure_console_show_filllog(GtkEditable *editable, gpointer data)
{
  GPtrArray *t=data;
  GtkWidget *log=GTK_WIDGET(conc_tuple_get(t,0));
  GtkWidget *nolog=GTK_WIDGET(conc_tuple_get(t,1));
  gchar *s;
  gint i;

  if (GTK_WIDGET_IS_SENSITIVE(log)) return;
  if (GTK_WIDGET_STATE(GTK_WIDGET(nolog))==GTK_STATE_ACTIVE) return;
  s=gtk_editable_get_chars(GTK_EDITABLE(editable),0,-1);
  gtk_editable_delete_text(GTK_EDITABLE(log),0,-1);
  i=0;
  gtk_editable_insert_text(GTK_EDITABLE(log),s,strlen(s),&i);
  g_free(s);
}


static void
conc_configure_console_show(GtkWidget *vbox, GtkWidget **name,
                            GtkWidget **server, GtkWidget **device,
                            GtkWidget **bits8, GtkWidget **bits7,
                            GtkWidget **stopbits1, GtkWidget **stopbits2,
                            GtkWidget **none, GtkWidget **odd, GtkWidget **even,
                            GtkWidget **transmit, GtkWidget **receive,
                            GtkWidget **log,
                            TermSpec *spec)
{
  GtkWidget *w,*x,*y,*n,*z,*u;
  int i,j;
  gboolean log_default=FALSE;
  gboolean log_empty=FALSE;

  if (!spec) spec=&conc_default_spec;
  *log=gtk_entry_new();
  z=gtk_radio_button_new_with_label(NULL,"do not log");
  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox),x,TRUE,TRUE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Name:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  *name=gtk_entry_new();
  if (spec->name) {
    gtk_entry_append_text(GTK_ENTRY(*name),spec->name);
  }
  gtk_signal_connect(GTK_OBJECT(*name), "changed",
                     GTK_SIGNAL_FUNC(conc_configure_console_show_filllog),
                     conc_tuple(*name,*log,z,NULL));
  gtk_box_pack_start(GTK_BOX(x),*name,TRUE,TRUE,0);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,0);

  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox),x,TRUE,TRUE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Server:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  *server=gtk_option_menu_new();
  w=gtk_menu_new();
  j=-1;
  for (i=0;i<conc_list_length(G->server_list);i++) {
    gtk_menu_append(GTK_MENU(w),gtk_menu_item_new_with_label(conc_list(G->server_list,i)));
    if (spec->server && !strcmp(spec->server,conc_list(G->server_list,i))) j=i;
  }
  gtk_option_menu_set_menu(GTK_OPTION_MENU(*server),w);
  if (j>-1) gtk_option_menu_set_history(GTK_OPTION_MENU(*server),j);
  gtk_box_pack_start(GTK_BOX(x),*server,TRUE,TRUE,0);
  if (spec!=&conc_default_spec) gtk_widget_set_sensitive(*server,FALSE);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,0);

  n=gtk_notebook_new();
  gtk_box_pack_start(GTK_BOX(vbox),n,TRUE,TRUE,GNOME_PAD_SMALL);
  y=gtk_vbox_new(FALSE,GNOME_PAD);
  gtk_notebook_append_page(GTK_NOTEBOOK(n),y,gtk_label_new(_("Serial Line")));

  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(y),x,TRUE,TRUE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Device:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  *device=gtk_entry_new();
  if (spec->device) {
    gtk_entry_append_text(GTK_ENTRY(*device),spec->device);
  }
  gtk_box_pack_start(GTK_BOX(x),*device,TRUE,TRUE,0);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,0);

  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(y),x,TRUE,TRUE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Transmit Rate:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  *transmit=gtk_option_menu_new();
  gtk_option_menu_set_menu(GTK_OPTION_MENU(*transmit),speed_menu());
  i=set_speed(spec->t_rate);
  if (i>-1) gtk_option_menu_set_history(GTK_OPTION_MENU(*transmit),i);
  gtk_box_pack_start(GTK_BOX(x),*transmit,TRUE,TRUE,0);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,0);

  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(y),x,TRUE,TRUE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Receive Rate:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  *receive=gtk_option_menu_new();
  gtk_option_menu_set_menu(GTK_OPTION_MENU(*receive),speed_menu());
  i=set_speed(spec->r_rate);
  if (i>-1) gtk_option_menu_set_history(GTK_OPTION_MENU(*receive),i);
  gtk_box_pack_start(GTK_BOX(x),*receive,TRUE,TRUE,0);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,0);

  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(y),x,TRUE,TRUE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Data bits:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  *bits8=gtk_radio_button_new_with_label(NULL,"8");
  *bits7=gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(*bits8),"7");
  w=gtk_hbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(w),GNOME_PAD_SMALL);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(w),GTK_BUTTONBOX_SPREAD);
  gtk_button_box_set_child_size(GTK_BUTTON_BOX(w),0,0);
  gtk_box_pack_start(GTK_BOX(x),w,TRUE,TRUE,0);
  gtk_box_pack_end(GTK_BOX(w),*bits8,FALSE,FALSE,0);
  gtk_box_pack_end(GTK_BOX(w),*bits7,FALSE,FALSE,0);
  if (spec->data_bits==8) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(*bits8),TRUE);
  if (spec->data_bits==7) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(*bits7),TRUE);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,0);

  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(y),x,TRUE,TRUE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Parity:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  *none=gtk_radio_button_new_with_label(NULL,"none");
  *odd=gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(*none),"odd");
  *even=gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(*none),"even");
  w=gtk_hbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(w),GNOME_PAD_SMALL);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(w),GTK_BUTTONBOX_SPREAD);
  gtk_button_box_set_child_size(GTK_BUTTON_BOX(w),0,0);
  gtk_box_pack_start(GTK_BOX(x),w,TRUE,TRUE,0);
  gtk_box_pack_end(GTK_BOX(w),*none,FALSE,FALSE,0);
  gtk_box_pack_end(GTK_BOX(w),*odd,FALSE,FALSE,0);
  gtk_box_pack_end(GTK_BOX(w),*even,FALSE,FALSE,0);
  if (spec->parity==PARITY_NONE) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(*none),TRUE);
  if (spec->parity==PARITY_ODD) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(*odd),TRUE);
  if (spec->parity==PARITY_EVEN) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(*even),TRUE);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,0);

  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(y),x,TRUE,TRUE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Stop bits:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  *stopbits1=gtk_radio_button_new_with_label(NULL,"1");
  *stopbits2=gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(*stopbits1),"2");
  w=gtk_hbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(w),GNOME_PAD_SMALL);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(w),GTK_BUTTONBOX_SPREAD);
  gtk_button_box_set_child_size(GTK_BUTTON_BOX(w),0,0);
  gtk_box_pack_start(GTK_BOX(x),w,TRUE,TRUE,0);
  gtk_box_pack_end(GTK_BOX(w),*stopbits1,FALSE,FALSE,0);
  gtk_box_pack_end(GTK_BOX(w),*stopbits2,FALSE,FALSE,0);
  if (spec->stop_bits==1) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(*stopbits1),TRUE);
  if (spec->stop_bits==2) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(*stopbits2),TRUE);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,0);

  y=gtk_vbox_new(FALSE,GNOME_PAD);
  gtk_notebook_append_page(GTK_NOTEBOOK(n),y,gtk_label_new(_("Logging")));

  if (!spec->logfile || !strlen(spec->logfile)) {
    log_empty=TRUE;
  }
  if (spec->logfile && spec->name && !strcmp(spec->logfile,spec->name)) {
    log_default=TRUE;
  }
  x=gtk_vbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(x),GNOME_PAD_SMALL);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(x),GTK_BUTTONBOX_SPREAD);
  gtk_box_pack_start(GTK_BOX(y),x,FALSE,FALSE,GNOME_PAD_SMALL);
  gtk_signal_connect(GTK_OBJECT(z), "clicked",
                     GTK_SIGNAL_FUNC(conc_configure_console_show_greylog),
                     conc_tuple(z,*log,*name,"no log",NULL));
  gtk_box_pack_end(GTK_BOX(x),z,FALSE,FALSE,0);
  w=gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(z),"log to default file");
  gtk_signal_connect(GTK_OBJECT(w), "clicked",
                     GTK_SIGNAL_FUNC(conc_configure_console_show_greylog),
                     conc_tuple(w,*log,*name,"default log",NULL));
  gtk_box_pack_end(GTK_BOX(x),w,FALSE,FALSE,0);
  u=gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(z),"log to file");
  gtk_signal_connect(GTK_OBJECT(u), "clicked",
                     GTK_SIGNAL_FUNC(conc_configure_console_show_log),
                     *log);
  gtk_box_pack_end(GTK_BOX(x),u,FALSE,FALSE,0);
  x=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(y),x,FALSE,FALSE,GNOME_PAD_SMALL);
  if (log_empty) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(z),TRUE);
  if (log_default) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w),TRUE);
  if (!log_default && !log_empty) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(u),TRUE);
  w=gtk_label_new(_("File name:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  gtk_box_pack_start(GTK_BOX(x),*log,TRUE,TRUE,0);
  w=gtk_label_new("");
  if (log_default || log_empty) gtk_widget_set_sensitive(*log,FALSE);
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,0);

}

static void
conc_configure_remove_console_done(ConcChannel *chan, gchar *line,
                                  gpointer cd, gpointer d)
{
  GtkWidget *w;
  gchar *s;
  gint i;

  if (line[0]=='o') {
    g_free(line);
    return;
  }
  for (i=0;line[i];i++); if (line[i]=='\r') line[i]='\n';
  s=g_strdup_printf("Error removing console:\n%s",line);
  g_free(line);
  w=gnome_message_box_new(s,GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
  g_free(s);
  gtk_widget_show_all(w);
  gnome_dialog_run_and_close(GNOME_DIALOG(w));
}

static void
conc_configure_remove_console(GtkButton *button,gpointer data)
{
  GtkWidget **a;
  gchar *s,*t;

  a=G->config_app_w;
  if (!gtk_clist_get_text(GTK_CLIST(a[CONFIG_console_list]),
                          G->config_app_console_selection,0,&t)) {
    g_warning("could not fetch data from clist");
    return;
  }
  s=g_strdup_printf("cdc%s",t);
  conc_channel_command(G->srv_chan,s,conc_configure_remove_console_done,NULL);
  g_free(s);
}

static int
conc_configure_console_close(GnomeDialog *dialog, gpointer data)
{
  GtkWidget **a=(GtkWidget **)data;

  gtk_widget_set_sensitive(a[CONFIG_CONSOLE_button],TRUE);
  g_free(a);
  return FALSE;
}

void
conc_configure_console_cancel_button(GtkButton *b, gpointer data)
{
  GtkWidget **a=(GtkWidget **)data;

  gnome_dialog_close(GNOME_DIALOG(a[CONFIG_CONSOLE_dialog]));
}

static guint
get_menu_index(GtkWidget *menu)
{
  GtkWidget *item;

  item=gtk_menu_get_active(GTK_MENU(menu));
  return g_list_index(GTK_MENU_SHELL(menu)->children,item);
}

void
conc_configure_console_done(ConcChannel *chan, gchar *line,
                                gpointer cd, gpointer d)
{
  GtkWidget *w;
  Console *c;
  gchar *s;
  gchar *msg;
  gboolean new=FALSE;

  if (cd) {
    msg=_("Error modifying console");
    g_free(cd);
  } else {
    msg=_("Error opening console");
    new=TRUE;
  }
  if (line[0]=='e') {
    s=g_strdup_printf("%s:\n%s",msg,&line[1]);
    w=gnome_message_box_new(s,GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
    g_free(s);
    gtk_widget_show_all(w);
    gnome_dialog_run_and_close(GNOME_DIALOG(w));
  }  else if (new) {
    c=g_new0(Console,1);
    c->name=g_strdup(&line[3]);
    c->id=getid(&line[1]);
    c->active=TRUE;
    conc_add_console(c);
    conc_save_prefs();
  }
  g_free(line);
}

void
conc_configure_console_ok_button(GtkButton *b, gpointer data)
{
  GtkWidget **a=(GtkWidget **)data;
  TermSpec t;
  gchar *s,*p;
  GtkWidget *w;
  GList *l;
  Console *c=NULL;

  t.name=t.server=t.device=t.logfile=NULL;
  t.name=gtk_editable_get_chars(GTK_EDITABLE(a[CONFIG_CONSOLE_name]),0,-1);
  t.device=gtk_editable_get_chars(GTK_EDITABLE(a[CONFIG_CONSOLE_device]),0,-1);
  t.logfile=gtk_editable_get_chars(GTK_EDITABLE(a[CONFIG_CONSOLE_log]),0,-1);
  if (GTK_WIDGET_STATE(GTK_WIDGET(a[CONFIG_CONSOLE_bits8]))==GTK_STATE_ACTIVE) t.data_bits=8;
  if (GTK_WIDGET_STATE(GTK_WIDGET(a[CONFIG_CONSOLE_bits7]))==GTK_STATE_ACTIVE) t.data_bits=7;
  if (GTK_WIDGET_STATE(GTK_WIDGET(a[CONFIG_CONSOLE_stopbits1]))==GTK_STATE_ACTIVE) t.stop_bits=1;
  if (GTK_WIDGET_STATE(GTK_WIDGET(a[CONFIG_CONSOLE_stopbits2]))==GTK_STATE_ACTIVE) t.stop_bits=2;
  if (GTK_WIDGET_STATE(GTK_WIDGET(a[CONFIG_CONSOLE_none]))==GTK_STATE_ACTIVE) t.parity=PARITY_NONE;
  if (GTK_WIDGET_STATE(GTK_WIDGET(a[CONFIG_CONSOLE_odd]))==GTK_STATE_ACTIVE) t.parity=PARITY_ODD;
  if (GTK_WIDGET_STATE(GTK_WIDGET(a[CONFIG_CONSOLE_even]))==GTK_STATE_ACTIVE) t.parity=PARITY_EVEN;
    t.server=g_strdup(conc_list(G->server_list,get_menu_index(gtk_option_menu_get_menu(GTK_OPTION_MENU(a[CONFIG_CONSOLE_server])))));
  t.t_rate=get_speed(get_menu_index(gtk_option_menu_get_menu(GTK_OPTION_MENU(a[CONFIG_CONSOLE_transmit]))));
  t.r_rate=get_speed(get_menu_index(gtk_option_menu_get_menu(GTK_OPTION_MENU(a[CONFIG_CONSOLE_receive]))));
  if (!strlen(t.name)) {
    g_free(t.name); g_free(t.server); g_free(t.device);
    w=gnome_message_box_new(_("Name required."),GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,GNOME_STOCK_BUTTON_CANCEL,NULL);
    gtk_widget_show_all(w);
    if (gnome_dialog_run_and_close(GNOME_DIALOG(w))==1) gnome_dialog_close(GNOME_DIALOG(a[CONFIG_CONSOLE_dialog]));
    return;
  }
  if (!strlen(t.device) || t.device[0]!='/') {
    g_free(t.name); g_free(t.server); g_free(t.device);
    w=gnome_message_box_new(_("Invalid device."),GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,GNOME_STOCK_BUTTON_CANCEL,NULL);
    gtk_widget_show_all(w);
    if (gnome_dialog_run_and_close(GNOME_DIALOG(w))==1) gnome_dialog_close(GNOME_DIALOG(a[CONFIG_CONSOLE_dialog]));
    return;
  }
  p=conc_termspec_from(&t);
  if (a[CONFIG_CONSOLE_spec]) {
    s=conc_termspec_from((TermSpec *)a[CONFIG_CONSOLE_spec]);
    if (!strcmp(s,p)) { /*nothing was changed*/
      gnome_dialog_close(GNOME_DIALOG(a[CONFIG_CONSOLE_dialog]));
      return;
    }
    g_free(s);
    for (l=G->consoles;l;l=g_list_next(l)) {
      c=(Console *)l->data;
      if (!strcmp(((TermSpec *)a[CONFIG_CONSOLE_spec])->name,c->name)) break;
    }
    s=g_strdup_printf("cec%02x%s",c->id,p);
  } else {
    s=g_strdup_printf("cnc%s",p);
  }
  g_free(p);
  g_free(t.name); g_free(t.server); g_free(t.device);
  conc_channel_command(G->srv_chan,s,conc_configure_console_done,a[CONFIG_CONSOLE_spec]);
  g_free(s);
  gnome_dialog_close(GNOME_DIALOG(a[CONFIG_CONSOLE_dialog]));
}

static void
conc_configure_console2(ConcChannel *chan,gchar *line,gpointer cd,gpointer d)
{
  GtkWidget *w;
  GtkWidget **a;
  gchar *title,*t;
  TermSpec *spec=NULL;

  
  if (line && line[0]=='e') {
    t=g_strdup_printf("%s\n%s",_("Error retrieving configuarion information:"),&line[1]);
    w=gnome_message_box_new(t,GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
    g_free(t);
    gnome_dialog_run_and_close(GNOME_DIALOG(w));
    return;
  }
  if (line) {
    spec=conc_termspec_to(&line[1]);
    g_free(line);
    title=g_strdup_printf("%s %s",_("Edit Console"),spec->name);
  } else {
    title=g_strdup(_("New Console"));
  }
  a=g_new0(GtkWidget *,CONFIG_CONSOLE_END);
  a[CONFIG_CONSOLE_spec]=(GtkWidget *)spec;
  a[CONFIG_CONSOLE_button]=GTK_WIDGET(cd);
  gtk_widget_set_sensitive(GTK_WIDGET(cd),FALSE);
  a[CONFIG_CONSOLE_dialog]=gnome_dialog_new(title,GNOME_STOCK_BUTTON_OK,GNOME_STOCK_BUTTON_CANCEL,NULL);
  g_free(title);
  gnome_dialog_set_parent(GNOME_DIALOG(a[CONFIG_CONSOLE_dialog]),GTK_WINDOW(G->app));
  gtk_signal_connect(GTK_OBJECT(a[CONFIG_CONSOLE_dialog]), "close",
                     GTK_SIGNAL_FUNC(conc_configure_console_close),
                     a);
  gnome_dialog_button_connect(GNOME_DIALOG(a[CONFIG_CONSOLE_dialog]),0,conc_configure_console_ok_button,a);
  gnome_dialog_button_connect(GNOME_DIALOG(a[CONFIG_CONSOLE_dialog]),1,conc_configure_console_cancel_button,a);

  conc_configure_console_show(GNOME_DIALOG(a[CONFIG_CONSOLE_dialog])->vbox,
                              &a[CONFIG_CONSOLE_name],
                              &a[CONFIG_CONSOLE_server],
                              &a[CONFIG_CONSOLE_device],
                              &a[CONFIG_CONSOLE_bits8],
                              &a[CONFIG_CONSOLE_bits7],
                              &a[CONFIG_CONSOLE_stopbits1],
                              &a[CONFIG_CONSOLE_stopbits2],
                              &a[CONFIG_CONSOLE_none],
                              &a[CONFIG_CONSOLE_odd],
                              &a[CONFIG_CONSOLE_even],
                              &a[CONFIG_CONSOLE_transmit],
                              &a[CONFIG_CONSOLE_receive],
                              &a[CONFIG_CONSOLE_log],
                              spec);


  gtk_widget_show_all(a[CONFIG_CONSOLE_dialog]);
}

static void
conc_configure_console(GtkButton *button,gpointer data)
{
  gchar *action=(gchar *)data;
  GtkWidget **a=G->config_app_w;
  gchar *t,*s;

  if (action[0]=='e') {
    if (!gtk_clist_get_text(GTK_CLIST(a[CONFIG_console_list]),
                            G->config_app_console_selection,0,&t)) {
      g_warning("could not fetch data from clist");
      return;
    }
    s=g_strdup_printf("cic%s",t);
    conc_channel_command(G->srv_chan,s,conc_configure_console2,button);
    g_free(s);
  } else {
    conc_configure_console2(NULL,NULL,button,NULL);
  }
}

static int
conc_configure_close(GnomeDialog *dialog, gpointer data)
{
  gtk_widget_set_sensitive(GTK_WIDGET(data),TRUE);
  G->config_app=NULL;
  g_free(G->config_app_w);
  return FALSE;
}

static void
conc_configure_server_select(GtkCList *clist, gint row, gint column,
                               GdkEventButton *event, gpointer user_data)
{
  GtkWidget **a;

  a=G->config_app_w;
  gtk_widget_set_sensitive(a[CONFIG_server_remove],TRUE);
  G->config_app_server_selection=row;
}

static void
conc_configure_server_unselect(GtkCList *clist, gint row, gint column,
                               GdkEventButton *event, gpointer user_data)
{
  GtkWidget **a;

  a=G->config_app_w;
  gtk_widget_set_sensitive(a[CONFIG_server_remove],FALSE);
}

static void
conc_configure_console_select(GtkCList *clist, gint row, gint column,
                               GdkEventButton *event, gpointer user_data)
{
  GtkWidget **a;

  a=G->config_app_w;
  gtk_widget_set_sensitive(a[CONFIG_console_remove],TRUE);
  gtk_widget_set_sensitive(a[CONFIG_console_edit],TRUE);
  G->config_app_console_selection=row;
}

static void
conc_configure_console_unselect(GtkCList *clist, gint row, gint column,
                               GdkEventButton *event, gpointer user_data)
{
  GtkWidget **a;

  a=G->config_app_w;
  gtk_widget_set_sensitive(a[CONFIG_console_remove],FALSE);
  gtk_widget_set_sensitive(a[CONFIG_console_edit],FALSE);
}

void
conc_configure(GtkWidget *widget, gpointer data)
{
  GtkWidget *w,*y,*x,*n;
  GtkWidget **a;

  a=G->config_app_w=g_new0(GtkWidget *,CONFIG_END);
  gtk_widget_set_sensitive(widget,FALSE);
  G->config_app=gnome_dialog_new(_("Conc Configuration"),GNOME_STOCK_BUTTON_CLOSE,NULL);
  gnome_dialog_set_parent(GNOME_DIALOG(G->config_app),GTK_WINDOW(G->app));
  gnome_dialog_set_close(GNOME_DIALOG(G->config_app),TRUE);
  gtk_signal_connect(GTK_OBJECT(G->config_app), "close",
                     GTK_SIGNAL_FUNC(conc_configure_close),
                     widget);

  n=gtk_notebook_new();
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(G->config_app)->vbox),n,TRUE,TRUE,GNOME_PAD);

  x=gtk_hbox_new(FALSE,GNOME_PAD);
  gtk_notebook_append_page(GTK_NOTEBOOK(n),x,gtk_label_new(_("Servers")));
  a[CONFIG_server_list]=gtk_clist_new(1);
  gtk_clist_set_column_auto_resize(GTK_CLIST(a[CONFIG_server_list]),0,TRUE);
  gtk_signal_connect(GTK_OBJECT(a[CONFIG_server_list]), "unselect-row",
                     GTK_SIGNAL_FUNC(conc_configure_server_unselect),
                     NULL);
  gtk_signal_connect(GTK_OBJECT(a[CONFIG_server_list]), "select-row",
                     GTK_SIGNAL_FUNC(conc_configure_server_select),
                     NULL);
  gtk_box_pack_start(GTK_BOX(x),a[CONFIG_server_list],TRUE,TRUE,GNOME_PAD_SMALL);
  y=gtk_vbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(y),GNOME_PAD_SMALL);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(y),GTK_BUTTONBOX_SPREAD);
  gtk_box_pack_end(GTK_BOX(x),y,FALSE,FALSE,GNOME_PAD_SMALL);
  w=gtk_button_new_with_label(_("New..."));
  gtk_signal_connect(GTK_OBJECT(w), "clicked",
                     GTK_SIGNAL_FUNC(conc_configure_new_server),
                     widget);
  gtk_box_pack_end(GTK_BOX(y),w,FALSE,FALSE,0);
/*
  w=gtk_button_new_with_label(_("Edit"));
  gtk_widget_set_sensitive(w,FALSE);
  gtk_box_pack_end(GTK_BOX(y),w,FALSE,FALSE,0);
*/
  a[CONFIG_server_remove]=gtk_button_new_with_label(_("Remove"));
  gtk_signal_connect(GTK_OBJECT(a[CONFIG_server_remove]), "clicked",
                     GTK_SIGNAL_FUNC(conc_configure_remove_server),
                     NULL);
  gtk_widget_set_sensitive(a[CONFIG_server_remove],FALSE);
  gtk_box_pack_end(GTK_BOX(y),a[CONFIG_server_remove],FALSE,FALSE,0);
  conc_configure_list_server_changed();

  x=gtk_hbox_new(FALSE,GNOME_PAD);
  gtk_notebook_append_page(GTK_NOTEBOOK(n),x,gtk_label_new(_("Consoles")));
  a[CONFIG_console_list]=gtk_clist_new(1);
  gtk_clist_set_column_auto_resize(GTK_CLIST(a[CONFIG_console_list]),0,TRUE);
  gtk_signal_connect(GTK_OBJECT(a[CONFIG_console_list]), "unselect-row",
                     GTK_SIGNAL_FUNC(conc_configure_console_unselect),
                     NULL);
  gtk_signal_connect(GTK_OBJECT(a[CONFIG_console_list]), "select-row",
                     GTK_SIGNAL_FUNC(conc_configure_console_select),
                     NULL);
  gtk_box_pack_start(GTK_BOX(x),a[CONFIG_console_list],TRUE,TRUE,GNOME_PAD_SMALL);
  y=gtk_vbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(y),GNOME_PAD_SMALL);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(y),GTK_BUTTONBOX_SPREAD);
  gtk_box_pack_end(GTK_BOX(x),y,FALSE,FALSE,GNOME_PAD_SMALL);
  w=gtk_button_new_with_label(_("New..."));
  gtk_signal_connect(GTK_OBJECT(w), "clicked",
                     GTK_SIGNAL_FUNC(conc_configure_console),
                     "new");
  gtk_box_pack_end(GTK_BOX(y),w,FALSE,FALSE,0);
  a[CONFIG_console_edit]=gtk_button_new_with_label(_("Edit..."));
  gtk_signal_connect(GTK_OBJECT(a[CONFIG_console_edit]), "clicked",
                     GTK_SIGNAL_FUNC(conc_configure_console),
                     "edit");
  gtk_widget_set_sensitive(a[CONFIG_console_edit],FALSE);
  gtk_box_pack_end(GTK_BOX(y),a[CONFIG_console_edit],FALSE,FALSE,0);
  a[CONFIG_console_remove]=gtk_button_new_with_label(_("Remove"));
  gtk_widget_set_sensitive(a[CONFIG_console_remove],FALSE);
  gtk_signal_connect(GTK_OBJECT(a[CONFIG_console_remove]), "clicked",
                     GTK_SIGNAL_FUNC(conc_configure_remove_console),
                     widget);
  gtk_box_pack_end(GTK_BOX(y),a[CONFIG_console_remove],FALSE,FALSE,0);
  conc_configure_list_console_changed();

  gtk_widget_show_all(G->config_app);
}
