/*
 This file is part of conc.

    Conc is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Conc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with conc; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gnome.h>
#include <gconf/gconf.h>
#include "conc.h"
#include "conc-ui.h"

#define FONT "-adobe-courier-medium-r-normal-*-*-140-*-*-m-*-iso8859-1"
#define SCROLLBACK 100

void
conc_save_prefs(void)
{
  GSList *cs;
  GList *l;
  Console *c;
  GError *e=NULL;

  cs=NULL;
  for (l=G->consoles;l;l=g_list_next(l)) {
    c=(Console *)l->data;
    cs=g_slist_append(cs,c->name);
  }
  gconf_engine_set_list(G->gce,"/apps/conc/ui/consoles",GCONF_VALUE_STRING,cs,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  g_slist_free(cs);
  gconf_engine_set_int(G->gce,"/apps/conc/ui/columns",G->columns,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  gconf_engine_set_string(G->gce,"/apps/conc/ui/server",G->server,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  gconf_engine_set_string(G->gce,"/apps/conc/ui/font",G->console_font,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  gconf_engine_set_int(G->gce,"/apps/conc/ui/scrollback",G->console_scrollback,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  gconf_engine_suggest_sync(G->gce,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
}

void
conc_load_prefs(void)
{
  GSList *cs,*l;
  Console *c;
  GError *e=NULL;

  cs=gconf_engine_get_list(G->gce,"/apps/conc/ui/consoles",GCONF_VALUE_STRING,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  for (l=cs;l;l=g_slist_next(l)) {
    c=g_new0(Console,1);
    c->id=-1;
    c->name=(gchar *)l->data;
    conc_add_console(c);
  }
  g_slist_free(cs);
  G->columns=gconf_engine_get_int(G->gce,"/apps/conc/ui/columns",&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  G->server=gconf_engine_get_string(G->gce,"/apps/conc/ui/server",&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  if (!G->columns) G->columns=2;
  G->console_font=gconf_engine_get_string(G->gce,"/apps/conc/ui/font",&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  if (!G->console_font) G->console_font=g_strdup(FONT);
  G->console_scrollback=gconf_engine_get_int(G->gce,"/apps/conc/ui/scrollback",&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  if (!G->console_scrollback) G->console_scrollback=SCROLLBACK;
}

void
conc_preferences_apply_button(GtkButton *b, gpointer data)
{
  GList *l,*cl;
  Console *c;
  gboolean rep_del;
  gchar *spin_res;
  gint i;

  /* hide any consoles that are no longer wanted */
  do {
    rep_del=FALSE;
    for (cl=G->consoles;cl;cl=g_list_next(cl)) {
      c=(Console *)cl->data;
      for (l=G->prefs_app_console_list;l;l=g_list_next(l)) {
        if (!strcmp(c->name,(gchar *)l->data)) break;
      }
      if (!l) {
        conc_remove_console(c);
        g_free(c->name);
        l=g_list_next(cl);
        if (!l) break; /* no more to do anyway */
        cl=g_list_previous(l);
        if (!cl) {
          /* only one item in the list - do it again */
         rep_del=TRUE;
         break;
        }
        g_free(c);
      }
    }
  } while (rep_del);
  /* add any new consoles */
  for (l=G->prefs_app_console_list;l;l=g_list_next(l)) {
    for (cl=G->consoles;cl;cl=g_list_next(cl)) {
      c=(Console *)cl->data;
      if (!strcmp(c->name,(gchar *)l->data)) break;
    }
    if (!cl) {
      c=g_new0(Console,1);
      c->name=g_strdup((gchar *)l->data);
      c->id=getid(&((gchar *)l->data)[-2]); /* yuk! */
      c->active=TRUE;
      conc_add_console(c);
    }
  }
  spin_res=gtk_editable_get_chars(GTK_EDITABLE(G->prefs_app_new_columns),0,-1);
  i=atoi(spin_res); g_free(spin_res);
  if (i!=G->columns) conc_columns_new(i);
  if (strcmp(G->console_font,gnome_font_picker_get_font_name(GNOME_FONT_PICKER(G->prefs_app_font)))) {
    g_free(G->console_font);
    G->console_font=g_strdup(gnome_font_picker_get_font_name(GNOME_FONT_PICKER(G->prefs_app_font)));
    /* FIXME: update any open consoles */
  }
  spin_res=gtk_editable_get_chars(GTK_EDITABLE(G->prefs_app_new_scrollback),0,-1);
  i=atoi(spin_res); g_free(spin_res);
  if (i!=G->console_scrollback) {
    G->console_scrollback=i;
    /* FIXME: update any open consoles */
  }
  conc_save_prefs();
}

void
conc_preferences_ok_button(GtkButton *b, gpointer data)
{
  conc_preferences_apply_button(b,b);
  gnome_dialog_close(GNOME_DIALOG(G->prefs_app));
}

void
conc_preferences_cancel_button(GtkButton *b, gpointer data)
{
  gnome_dialog_close(GNOME_DIALOG(G->prefs_app));
}

void
conc_preferences_console_toggled(GtkToggleButton *button, gpointer data)
{
  gint i=(gint)data;

  if (GTK_WIDGET_STATE(GTK_WIDGET(button))==GTK_STATE_ACTIVE) {
    G->prefs_app_console_list=g_list_append(G->prefs_app_console_list,
                                            &G->prefs_app_v[i][2]);
  } else {
    G->prefs_app_console_list=g_list_remove(G->prefs_app_console_list,
                                            &G->prefs_app_v[i][2]);
  }
}

static int
conc_preferences_close(GnomeDialog *dialog, gpointer data)
{
  gtk_widget_set_sensitive(GTK_WIDGET(data),TRUE);
  G->prefs_app=NULL;
  g_strfreev(G->prefs_app_v);
  return FALSE;
}

void
conc_preferences(GtkWidget *widget, gpointer data)
{
  GtkWidget *n,*w,*x,*y,*v;
  GtkAdjustment *adj;
  gint i;
  GList *l;
  Console *c;

  gtk_widget_set_sensitive(widget,FALSE);
  G->prefs_app=gnome_dialog_new(_("Preferences"),GNOME_STOCK_BUTTON_OK,GNOME_STOCK_BUTTON_APPLY,GNOME_STOCK_BUTTON_CLOSE,NULL);
  gnome_dialog_set_parent(GNOME_DIALOG(G->prefs_app),GTK_WINDOW(G->app));
  gtk_signal_connect(GTK_OBJECT(G->prefs_app), "close",
                     GTK_SIGNAL_FUNC(conc_preferences_close),
                     widget);
  gnome_dialog_button_connect(GNOME_DIALOG(G->prefs_app),0,conc_preferences_ok_button,NULL);
  gnome_dialog_button_connect(GNOME_DIALOG(G->prefs_app),1,conc_preferences_apply_button,NULL);
  gnome_dialog_button_connect(GNOME_DIALOG(G->prefs_app),2,conc_preferences_cancel_button,NULL);
  gtk_window_set_policy(GTK_WINDOW(G->prefs_app),FALSE,TRUE,FALSE);

  n=gtk_notebook_new();
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(G->prefs_app)->vbox),n,TRUE,TRUE,GNOME_PAD);

  
  v=gtk_vbox_new(FALSE,GNOME_PAD);
  w=gtk_scrolled_window_new(NULL,NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(w),GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start(GTK_BOX(v),w,TRUE,TRUE,GNOME_PAD_SMALL);
  gtk_notebook_append_page(GTK_NOTEBOOK(n),v,gtk_label_new(_("Monitor Consoles")));
  y=gtk_vbox_new(FALSE,GNOME_PAD_SMALL);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(w),y);
  G->prefs_app_console_list=NULL;
  if (G->console_list) {
    G->prefs_app_v=g_new(gchar *,conc_list_length(G->console_list)+1);
    for (i=0;i<conc_list_length(G->console_list);i++) {
      G->prefs_app_v[i]=g_strdup(conc_list(G->console_list,i));
    }
    G->prefs_app_v[i]=NULL;
    for (l=G->consoles;l;l=g_list_next(l)) {
      c=(Console *)l->data;
      for (i=0;G->prefs_app_v[i];i++) {
        if (!strcmp(c->name,&G->prefs_app_v[i][2])) break;
      }
      if (!G->prefs_app_v[i]) continue; /* this console no longer exists */
      G->prefs_app_console_list=g_list_append(G->prefs_app_console_list,
                                              &G->prefs_app_v[i][2]);
    }
    for (i=0;G->prefs_app_v[i];i++) {
      w=gtk_check_button_new_with_label(&G->prefs_app_v[i][2]);
      for (l=G->prefs_app_console_list;l;l=g_list_next(l)) {
        if (l->data==&G->prefs_app_v[i][2]) break;
      }
      if (l) {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w),TRUE);
      }
      gtk_box_pack_start(GTK_BOX(y),w,FALSE,FALSE,0);
      gtk_signal_connect(GTK_OBJECT(w),"toggled",
                         GTK_SIGNAL_FUNC(conc_preferences_console_toggled),
                         (gpointer)i);
    }
  }
  adj=(GtkAdjustment *)gtk_adjustment_new((gfloat)G->columns,1.0,10.0,1.0,1.0,1.0);
  x=gtk_hbox_new(FALSE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Columns:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  G->prefs_app_new_columns=gtk_spin_button_new(adj,1.0,0);
  gtk_box_pack_start(GTK_BOX(x),G->prefs_app_new_columns,
                     FALSE,FALSE,GNOME_PAD_SMALL);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,GNOME_PAD_SMALL);
  gtk_box_pack_start(GTK_BOX(v),x,FALSE,FALSE,GNOME_PAD_SMALL);
    
  v=gtk_vbox_new(FALSE,GNOME_PAD);
  gtk_notebook_append_page(GTK_NOTEBOOK(n),v,gtk_label_new(_("Terminal")));

  x=gtk_hbox_new(FALSE,GNOME_PAD_SMALL);
  gtk_box_pack_start(GTK_BOX(v),x,FALSE,FALSE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Font:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  G->prefs_app_font=gnome_font_picker_new();
  gnome_font_picker_set_mode(GNOME_FONT_PICKER(G->prefs_app_font),GNOME_FONT_PICKER_MODE_FONT_INFO);
  gnome_font_picker_set_font_name(GNOME_FONT_PICKER(G->prefs_app_font),G->console_font);
  gtk_box_pack_start(GTK_BOX(x),G->prefs_app_font,
                     FALSE,FALSE,GNOME_PAD_SMALL);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,GNOME_PAD_SMALL);

  x=gtk_hbox_new(FALSE,GNOME_PAD_SMALL);
  gtk_box_pack_start(GTK_BOX(v),x,FALSE,FALSE,GNOME_PAD_SMALL);
  w=gtk_label_new(_("Scrollback lines:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  adj=(GtkAdjustment *)gtk_adjustment_new((gfloat)G->console_scrollback,100.0,100000.0,1.0,5.0,1.0);
  G->prefs_app_new_scrollback=gtk_spin_button_new(adj,1.0,0);
  gtk_box_pack_start(GTK_BOX(x),G->prefs_app_new_scrollback,
                     FALSE,FALSE,GNOME_PAD_SMALL);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,GNOME_PAD_SMALL);

  gtk_widget_show_all(G->prefs_app);
}
