/*
 This file is part of conc.

    Conc is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Conc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with conc; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <glib.h>
#include <gconf/gconf.h>
#include <popt.h>
#include "conc.h"

#define LISTEN_QUEUE 8
#define LOGFILE "conc.log"
#define USAGE ""
#define HELP "Run '%s --help' to see a full list of command line arguments\n"

GMainLoop *gml;

GConfEngine *gce;

guchar idcount=0;

gboolean debug=FALSE;

typedef struct _Client Client;

typedef struct {
  ConcChannel *chan;
  gboolean active;
  gchar *name;
  gchar *termserv;
  gboolean local;
  guint timeout_id;
  gint is_new;
  Client *creator;
} Ts;
GList *tss=NULL;

typedef struct {
  gboolean short_to_set;
  gboolean long_to_set;
  guint short_to_id;
  guint long_to_id;
  time_t last_active;
  gchar status;
} Am;

typedef struct {
  gint id;
  TermSpec spec;
  Ts *ts;
  gboolean active;
  GString *log;
  guint timeout_id;
  guint connected_clients;
  Am am;
  gint is_new;
  Client *creator;
} Term;
GList *terms=NULL;

struct _Client {
  gchar *name;
  ConcChannel *chan;
  GList *terms;
  GList *logs;
  GList *ams;
  gboolean recv_config_change;
};
GList *clients=NULL;

typedef struct {
  Term *t;
  gchar *s;
} activatetermContainer;

typedef struct {
  Term *t;
  Client *c;
  gint command_id;
} configurationremoveconsoleContainer;

typedef struct {
  Client *c;
  GString *s;
  gint id;
  gint fd;
  gint errfd;
  gint command_id;
} logreaderContainer;

typedef struct {
  Client *c;
  gint id;
} resettermContainer;

typedef struct {
  Client *c;
  Term *t;
  gint id;
  TermSpec *spec;
} edittermContainer;


static gboolean conc_wait_do(gpointer data);
static void conc_wait(void);
static void activateterm2(ConcChannel *chan, gchar *line, gpointer cd, gpointer d);
static gboolean activateterm(gpointer data);
static gint gethex(gchar c);
static gint getid(gchar *p);
static void sendterm(gint id,gchar byte);
static void send_active(Term *t,gchar action);
static void send_console_users(Term *t);
static gboolean conc_term_idle(gpointer data);
static gboolean conc_term_tooidle(gpointer data);
static void log(Term *t);
static void readterm(gint id,gchar byte);
static void ts_closed(ConcChannel *chan, gpointer data);
static void readts(ConcChannel *chan, gchar *line, gint id, gpointer data);
static gboolean activatets(gpointer data);
static void save_config(void);
static void load_config(void);
static void closeclient(ConcChannel *chan, gpointer data);
static void listterms(Client *cl, gint command_id);
static void sendbreak(gint id);
static void resetterm_done(ConcChannel *chan, gchar *line, gpointer cd, gpointer d);
static void resetterm(Client *c,gint id,gint command_id);
static void registerterm(Client *c,gchar type,gint id);
static void deregisterterm(Client *c,gchar type,gint id);
static void configuration_change(gchar *s);
static void configuration_remove_console(ConcChannel *chan, gchar *line, gpointer cd, gpointer d);
static void editterm1(ConcChannel *chan,gchar *line,gpointer cd, gpointer d);
static void configuration_edit(Client *c,gchar *line, gint command_id);
static void configuration_info(Client *c,gchar *line, gint command_id);
static void configuration_delete(Client *c,gchar *line, gint command_id);
static void configuration_new(Client *c,gchar *line, gint command_id);
static void configuration(Client *c,gchar *line, gint command_id);
static gboolean log_reader(gpointer data);
static void list_console_users(Client *cl,gint id,gint command_id);
static void read_log(Client *c,gint id,gint lines,gint command_id);
static void readclient(ConcChannel *chan, gchar *line, gint command_id, gpointer data);
static void client_name(ConcChannel *chan,gchar *line,gpointer cd, gpointer d);
static gboolean new_client(GIOChannel *chan, GIOCondition cond, gpointer data);
static void start_service(void);
static void check_config(void);
static void conc_log(const gchar *domain, GLogLevelFlags level, const gchar *m, gpointer d);

static gboolean
conc_wait_do(gpointer data)
{
  switch(waitpid(0,NULL,WNOHANG)) {
    case 0:
      return TRUE;
    case -1:
      g_warning("wait syscall failed");
      return FALSE;
    default:
      return FALSE;
  }
}

static void
conc_wait(void)
{
  g_timeout_add(5000,conc_wait_do,NULL);
}

static void
activateterm2(ConcChannel *chan, gchar *line, gpointer cd, gpointer d)
{
  activatetermContainer *a=(activatetermContainer *)cd;
  Term *t=a->t;
  gchar *s=a->s;

  g_free(s);
  g_free(a);
  if (line[0]=='o') {
    t->active=TRUE;
    if (t->is_new>-1) {
      s=g_strdup_printf("cnc%02x%s",t->id,t->spec.name);
      configuration_change(s);
      g_free(s);
      s=g_strdup_printf("o%02x%s",t->id,t->spec.name);
      conc_channel_respond(t->creator->chan,s,t->is_new);
      g_free(s);
      t->is_new=FALSE;
      save_config();
    }
  } else {
    if (t->is_new>-1) {
      conc_channel_respond(t->creator->chan,line,t->is_new);
      terms=g_list_remove(terms,t);
      g_free(t->spec.name);
      g_free(t->spec.server);
      g_free(t->spec.device);
      g_string_free(t->log,TRUE);
      g_free(t);
    }
    /*FIXME: needs logging...*/
    g_warning("%s",line);
  }
}

static gboolean
activateterm(gpointer data)
{
  Term *t=(Term *)data;
  gchar *s,*p;
  activatetermContainer *a=g_new(activatetermContainer,1);

  if (t->active) return FALSE;
  if (!t->ts->active) return TRUE;
  p=conc_termspec_from(&t->spec);
  s=g_strdup_printf("n%02x%s",t->id,p);
  g_free(p);
  a->t=t; a->s=s;
  conc_channel_command(t->ts->chan,s,activateterm2,a);
  return FALSE;
}

static gint
gethex(gchar c)
{
  switch (c) {
    case '0': case '1': case '2': case '3': case '4': case '5': case '6':
    case '7': case '8': case '9': return(c-'0');
    case 'a': case 'A': return 10;
    case 'b': case 'B': return 11;
    case 'c': case 'C': return 12;
    case 'd': case 'D': return 13;
    case 'e': case 'E': return 14;
    case 'f': case 'F': return 15;
  }
  return 0;
}

static gint
getid(gchar *p)
{
  return gethex(p[0])*16+gethex(p[1]);
}

static void
sendterm(gint id,gchar byte)
{
  GList *l;
  Term *t;
  gchar *s;

  for (l=terms;l;l=g_list_next(l)) {
    t=(Term *)l->data;
    if (id==t->id) break;
  }
  if (!l) {
    g_warning("client attempted send to non existant term id %02x",id);
    return;
  }
  switch (byte) {
    case 0:
      s=g_strdup_printf("z%02x",id);
      break;
    case '\n':
      s=g_strdup_printf("T%02x",id);
      break;
    default:
      s=g_strdup_printf("t%02x%c",id,byte);
      break;
  }
  conc_channel_write(t->ts->chan,s);
  g_free(s);
}

static void
send_active(Term *t,gchar action)
{
  gchar *s;
  GList *cl,*l;
  Client *c;
  time_t now;

  now=time(NULL);
  if (action && action==t->am.status && !(action=='a' && now-t->am.last_active>4)) return;
  if (action=='a') t->am.last_active=now;
  if (!action) action='i';
  t->am.status=action;
  s=g_strdup_printf("a%02x%c%s",t->id,action,get_time(&t->am.last_active));
  for (cl=clients;cl;cl=g_list_next(cl)) {
    c=(Client *)cl->data;
    for (l=c->ams;l;l=g_list_next(l)) {
      if (t->id==(gint)l->data) {
        conc_channel_write(c->chan,s);
      }
    }
  }
  g_free(s);
}

static void
send_console_users(Term *t)
{
  GList *l;
  Client *c;
  gchar *s;

  s=g_strdup_printf("U%02x%d",t->id,t->connected_clients);
  for (l=clients;l;l=g_list_next(l)) {
    c=(Client *)l->data;
    if (g_list_find(c->terms,GINT_TO_POINTER(t->id))) {
      conc_channel_write(c->chan,s);
    }
  }
  g_free(s);
}

static gboolean
conc_term_idle(gpointer data)
{
  Term *t=(Term *)data;

  t->am.short_to_set=FALSE;
  send_active(t,'i');
  return FALSE;
}

static gboolean
conc_term_tooidle(gpointer data)
{
  Term *t=(Term *)data;

  t->am.long_to_set=FALSE;
  send_active(t,'x');
  return FALSE;
}

static void
log(Term *t)
{
  gint fd;
  gchar *fn;

  if (t->spec.logfile[0]=='/') {
    fn=g_strdup(t->spec.logfile);
  } else {
    fn=g_strdup_printf("%s%s",LOG_DIR,t->spec.logfile);
  }
  fd=open(fn,O_WRONLY|O_CREAT|O_APPEND,S_IRWXU);
  write(fd,get_time(NULL),19);
  write(fd," ",1);
  write(fd,t->log->str,t->log->len);
  close(fd);
  g_free(fn);
}

static void
readterm(gint id,gchar byte)
{
  GList *cl;
  Client *c=NULL;
  GList *l;
  Term *t=NULL;
  gchar *s;

  switch (byte) {
    case 0:
      s=g_strdup_printf("z%02x",id);
      break;
    case '\n':
      s=g_strdup_printf("T%02x",id);
      break;
    default:
      s=g_strdup_printf("t%02x%c",id,byte);
      break;
  }
  for (cl=clients;cl;cl=g_list_next(cl)) {
    c=(Client *)cl->data;
    for (l=c->terms;l;l=g_list_next(l)) {
      if (id==(gint)l->data) {
        conc_channel_write(c->chan,s);
      }
    }
  }
  g_free(s);
  for (l=terms;l;l=g_list_next(l)) {
    t=(Term *)l->data;
    if (id!=t->id) continue;
    if (!t->log) t->log=g_string_new("");
    if (byte=='\n') {
      t->log=conc_vtstrip(t->log);
      s=g_strdup_printf("l%02x%s",id,t->log->str);
      for (cl=clients;cl;cl=g_list_next(cl)) {
        c=(Client *)cl->data;
        for (l=c->logs;l;l=g_list_next(l)) {
          if (id==(gint)l->data) {
            conc_channel_write(c->chan,s);
          }
        }
      }
      t->log=g_string_append_c(t->log,'\n');
      if (strlen(t->spec.logfile)) log(t);
      g_string_free(t->log,TRUE);
      t->log=NULL;
      g_free(s);
      break;
    }
    t->log=g_string_append_c(t->log,byte);
    break;
  }
  if (t->am.short_to_set) g_source_remove(t->am.short_to_id);
  if (t->am.long_to_set) g_source_remove(t->am.long_to_id);
  t->am.short_to_set=TRUE;
  t->am.short_to_id=g_timeout_add(5000,conc_term_idle,t);
  t->am.long_to_set=TRUE;
  t->am.long_to_id=g_timeout_add(1000*60*60*24,conc_term_tooidle,t);
  send_active(t,'a');
}

static void
ts_closed(ConcChannel *chan, gpointer data)
{
  Ts *ts=(Ts *)data;
  gchar *s;

  if (!ts->active) {
    if (ts->is_new!=-1) {
      conc_channel_respond(ts->creator->chan,"efailed to start server",ts->is_new);
      ts->is_new=-1;
    }
    /*closed by user*/
    tss=g_list_remove(tss,ts);
    s=g_strdup_printf("crs%s",ts->name);
    g_free(ts->name);
    g_free(ts);
    configuration_change(s);
    g_free(s);
    save_config();
  } else {
    g_warning("termserv on %s died",ts->name);
    ts->active=FALSE;
    ts->timeout_id=g_timeout_add(500,activatets,ts);
  }
  conc_wait();
}

static void
readts(ConcChannel *chan, gchar *line, gint id, gpointer data)
{
  Ts *ts=(Ts *)data;

  switch (line[0]) {
    case 'z': /*recieved null*/
      readterm(getid(&line[1]),0);
      break;
    case 't': /*recived character*/
      readterm(getid(&line[1]),line[3]);
      break;
    case 'T':
      readterm(getid(&line[1]),'\n');
      break;
    case 'e':
      g_warning("error from server %s: %s",ts->name,&line[1]);
      break;
  }
  g_free(line);
}

static gboolean
activatets(gpointer data)
{
  Ts *ts=(Ts *)data;
  gint tofd[2], fromfd[2], errfd[2];
  gchar *s;
  gchar errbuf[256];
  gint i;

  if (ts->active) return FALSE;
  g_message("starting termserv on %s",ts->name);
  pipe(tofd); pipe(fromfd); pipe(errfd);
  switch (fork()) {
    case -1:
      g_warning("could not fork!");
      if (ts->is_new!=-1) {
        conc_channel_respond(ts->creator->chan,"efailed to fork",ts->is_new);
        tss=g_list_remove(tss,ts);
        g_free(ts->name);
        g_free(ts);
        return FALSE;
      }
      return TRUE;
    case 0:
      close(0); close(1); close(2);
      close(tofd[1]); close(fromfd[0]); close(errfd[0]);
      dup2(tofd[0],0); dup2(fromfd[1],1); dup2(errfd[1],2);
      if (ts->local) {
        execlp("termserv",NULL);
        g_assert_not_reached();
      } else {
        execlp("rsh",ts->name,ts->termserv,NULL);
        g_assert_not_reached();
      }
  }
  close(tofd[0]); close(fromfd[1]); close(errfd[1]);
  ts->chan=conc_channel_new(tofd[1],fromfd[0],ts_closed,readts,ts,CONC_PASSWORD);
  if (!ts->chan) {
    fcntl(errfd[0],F_SETFL,O_NONBLOCK);
    i=read(errfd[0],&errbuf[1],254);
    if (i>0) {
      errbuf[i+1]='\0';
      errbuf[0]='e';
      for (i=0;errbuf[i];i++) {
        if (errbuf[i]=='\n') {
          errbuf[i]='\0';
          break;
        }
      }
      g_warning("failed to start termserv %s: %s",ts->name,errbuf);
    } else {
      g_warning("failed to start termserv %s",ts->name);
    }
    if (ts->is_new!=-1) {
      if (i>0) {
        conc_channel_respond(ts->creator->chan,errbuf,ts->is_new);
      } else {
        conc_channel_respond(ts->creator->chan,"efailed to connect",ts->is_new);
      }
      tss=g_list_remove(tss,ts);
      g_free(ts->name);
      g_free(ts);
      return FALSE;
    } else {
      return TRUE;
    }
  }
  ts->active=TRUE;
  if (ts->is_new!=-1) {
    conc_channel_respond(ts->creator->chan,"ok",ts->is_new);
    ts->is_new=-1;
    s=g_strdup_printf("cns%s",ts->name);
    configuration_change(s);
    g_free(s);
    save_config();
  }
  return FALSE;
}

static void
save_config(void)
{
  Ts *ts;
  Term *t;
  GSList *list,*sl;
  GList *l;
  GError *e=NULL;
  gchar *s;

  list=NULL;
  for (l=tss;l;l=g_list_next(l)) {
    ts=(Ts *)l->data;
    s=g_strdup_printf("%s\r%s",ts->name,ts->termserv);
    list=g_slist_append(list,s);
  }
  gconf_engine_set_list(gce,"/apps/conc/server/servers",GCONF_VALUE_STRING,list,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  for (sl=list;sl;sl=g_slist_next(sl)) g_free(sl->data);
  g_slist_free(list);

  list=NULL;
  for (l=terms;l;l=g_list_next(l)) {
    t=(Term *)l->data;
    list=g_slist_append(list,conc_termspec_from(&t->spec));
  }
  gconf_engine_set_list(gce,"/apps/conc/server/consoles",GCONF_VALUE_STRING,list,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  for (sl=list;sl;sl=g_slist_next(sl)) g_free(sl->data);
  g_slist_free(list);

  gconf_engine_suggest_sync(gce,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
}

static void
load_config(void)
{
  Ts *ts;
  Term *t;
  GSList *list,*l;
  GList *sl;
  GError *e=NULL;
  TermSpec *spec;
  gchar **v;

  g_message("loading configuration");
  list=gconf_engine_get_list(gce,"/apps/conc/server/servers",GCONF_VALUE_STRING,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  for (l=list;l;l=g_slist_next(l)) {
    ts=g_new0(Ts,1);
    ts->is_new=-1;
    v=g_strsplit((gchar *)l->data,"\r",0);
    g_free(l->data);
    ts->name=g_strdup(v[0]);
    ts->termserv=g_strdup(v[1]);
    g_strfreev(v);
    g_message("termserv %s (%s)",ts->name,ts->termserv);
    if (!strcmp(ts->name,"localhost")) ts->local=TRUE;
    tss=g_list_append(tss,ts);
    ts->timeout_id=g_timeout_add(500,activatets,ts);
  }
  g_slist_free(list);

  list=gconf_engine_get_list(gce,"/apps/conc/server/consoles",GCONF_VALUE_STRING,&e);
  if (e) {
    g_warning("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  for (l=list;l;l=g_slist_next(l)) {
    t=g_new0(Term,1);
    spec=conc_termspec_to((gchar *)l->data);
    g_message("console %s on %s",spec->name,spec->server);
    g_free(l->data);
    t->spec=*spec; g_free(spec);
    t->id=idcount++;
    t->is_new=-1;
    for (sl=tss;sl;sl=g_list_next(sl)) {
      ts=(Ts *)sl->data;
      if (!strcmp(ts->name,t->spec.server)) break;
    }
    if (sl) {
      t->ts=ts;
      terms=g_list_append(terms,t);
      t->timeout_id=g_timeout_add(500,activateterm,t);
    } else {
      g_warning("no server %s, disabling console for %s",t->spec.server,t->spec.name);
      g_free(t->spec.name);
      g_free(t->spec.server);
      g_free(t->spec.device);
      g_free(t->spec.logfile);
      g_free(t);
    }
  }
  g_slist_free(list);

  g_message("configuration complete");

}

static void
closeclient(ConcChannel *chan, gpointer data)
{
  Client *c=(Client *)data;
  GList *l,*m;
  Term *t;

  g_message("client %s closed",c->name?c->name:"<unknown>");
  for (l=c->terms;l;l=g_list_next(l)) {
    for (m=terms;m;m=g_list_next(m)) {
      t=(Term *)m->data;
      if (GPOINTER_TO_INT(l->data)!=t->id) continue;
      t->connected_clients--;
      send_console_users(t);
      continue;
    }
  }
  l=g_list_find(clients,c);
  clients=g_list_remove_link(clients,l);
  g_list_free(c->terms);
  g_list_free(c->logs);
  g_list_free(l);
}

static void
listterms(Client *cl, gint command_id)
{
  GList *l;
  Term *t;
  GString *s;
  gchar *s2;
  gboolean got_one=FALSE;

  s=g_string_new("d");
  for (l=terms;l;l=g_list_next(l)) {
    t=(Term *)l->data;
    s2=g_strdup_printf("%02x%s:",t->id,t->spec.name);
    s=g_string_append(s,s2);
    g_free(s2);
    got_one=TRUE;
  }
  if (got_one) s=g_string_truncate(s,s->len-1);
  conc_channel_respond(cl->chan,s->str,command_id);
  g_string_free(s,TRUE);
}

static void
sendbreak(gint id)
{
  GList *l;
  Term *t;
  gchar *s;

  for (l=terms;l;l=g_list_next(l)) {
    t=(Term *)l->data;
    if (id==t->id) break;
  }
  if (!l) {
    g_warning("client attempted send to non existant term id %02x",id);
    return;
  }
  s=g_strdup_printf("b%02x",id);
  conc_channel_write(t->ts->chan,s);
  g_free(s);
}

static void
resetterm_done(ConcChannel *chan, gchar *line, gpointer cd, gpointer d)
{
  resettermContainer *a=(resettermContainer *)cd;

  /* FIXME: what if the client closed? */
  conc_channel_respond(a->c->chan,line,a->id);
  /* FIXME: if there is an error, the console must be closed */
  g_free(a);
}

static void
resetterm(Client *c,gint id,gint command_id)
{
  GList *l;
  Term *t;
  gchar *s;
  resettermContainer *a;

  for (l=terms;l;l=g_list_next(l)) {
    t=(Term *)l->data;
    if (id==t->id) break;
  }
  if (!l) {
    g_warning("client attempted reset to non existant term id %02x",id);
    return;
  }
  a=g_new0(resettermContainer,1);
  a->id=command_id;
  a->c=c;
  s=g_strdup_printf("r%02x",id);
  conc_channel_command(t->ts->chan,s,resetterm_done,a);
  g_free(s);
}

static void
registerterm(Client *c,gchar type,gint id)
{
  GList *l;
  Term *t;

  switch (type) {
    case 't': /* terminal */
      for (l=terms;l;l=g_list_next(l)) {
        t=(Term *)l->data;
        if (t->id!=id) continue;
        t->connected_clients++;
        break;
      }
      if (!l) break; /* FIXME: report error */
      c->terms=g_list_append(c->terms,(gpointer)id);
      send_console_users(t);
      break;
    case 'l': /* log */
      c->logs=g_list_append(c->logs,(gpointer)id);
      break;
    case 'a': /* activity */
      c->ams=g_list_append(c->ams,(gpointer)id);
      for (l=terms;l;l=g_list_next(l)) {
        t=(Term *)l->data;
        if (t->id!=id) continue;
        type=t->am.status;
        t->am.status='\0';
        send_active(t,type);
        break;
      }
      break;
    case 'c': /* configuration changes */
      c->recv_config_change=TRUE;
      break;
  }
}

static void
deregisterterm(Client *c,gchar type,gint id)
{
  GList *l;
  Term *t;

  switch (type) {
    case 't': /* terminal */
      for (l=terms;l;l=g_list_next(l)) {
        t=(Term *)l->data;
        if (t->id!=id) continue;
        t->connected_clients--;
        break;
      }
      if (!l) break; /* FIXME: report error */
      c->terms=g_list_remove(c->terms,(gpointer)id);
      break;
    case 'l': /* log */
      c->logs=g_list_remove(c->logs,(gpointer)id);
      break;
    case 'a': /* activity */
      c->ams=g_list_remove(c->ams,(gpointer)id);
      break;
    case 'c': /* configuration changes */
      c->recv_config_change=FALSE;
      break;
  }
}

static void
configuration_change(gchar *s)
{
  GList *l;
  Client *c;

  for (l=clients;l;l=g_list_next(l)) {
    c=(Client *)l->data;
    if (c->recv_config_change) {
      conc_channel_write(c->chan,s);
    }
  }
}

static void
configuration_remove_console(ConcChannel *chan, gchar *line,
                             gpointer cd, gpointer d)
{
  configurationremoveconsoleContainer *a=(configurationremoveconsoleContainer *)cd;
  gchar *s;

  if (line[0]!='o') {
    conc_channel_respond(a->c->chan,line,a->command_id);
    g_free(a);
    return;
  }
  terms=g_list_remove(terms,a->t);
  s=g_strdup_printf("crc%s",a->t->spec.name);
  g_free(a->t->spec.name);
  g_free(a->t->spec.server);
  g_free(a->t->spec.device);
  if (a->t->log) g_string_free(a->t->log,TRUE);
  g_free(a->t);
  configuration_change(s);
  g_free(s);
  conc_channel_respond(a->c->chan,"ok",a->command_id);
  g_free(a);
  save_config();
}

static void
editterm1(ConcChannel *chan,gchar *line,gpointer cd, gpointer d)
{
  edittermContainer *a=(edittermContainer *)cd;
  gchar *s,*t;

  if (line[0]=='o') {
    s=a->t->spec.name;
    g_free(a->t->spec.server);
    g_free(a->t->spec.device);
    a->t->spec=*a->spec;
    conc_channel_respond(a->c->chan,"ok",a->id);
    if (strcmp(s,a->t->spec.name)) {
      t=g_strdup_printf("ccc%s\r%s",s,a->t->spec.name);
      configuration_change(t);
      g_free(t);
    }
    g_free(s);
    save_config();
  } else {
    /* FIXME: terminal is now closed! */
    g_free(a->spec->name);
    g_free(a->spec->server);
    g_free(a->spec->device);
    conc_channel_respond(a->c->chan,line,a->id);
  }
  g_free(a->spec);
  g_free(a);
}

static void
configuration_edit(Client *c,gchar *line, gint command_id)
{
  GList *l;
  Term *t;
  gchar *s;
  gint i;
  edittermContainer *a;
  TermSpec *spec;

  switch (line[2]) {
    case 'c': /* edit console  - requires response*/
      i=getid(&line[3]);
      spec=conc_termspec_to(&line[5]);
      for (l=terms;l;l=g_list_next(l)) {
        t=(Term *)l->data;
        if (i!=t->id && !strcmp(t->spec.name,spec->name)) break;
      }
      if (l) {
        g_free(spec->name);
        g_free(spec->server);
        g_free(spec->device);
        g_free(spec);
        s=g_strdup_printf("ethere is already a console named %s",t->spec.name);
        conc_channel_respond(c->chan,s,command_id);
        g_free(s);
        break;
      }
      for (l=terms;l;l=g_list_next(l)) {
        t=(Term *)l->data;
        if (i==t->id) break;
      }
      if (!l) {
        conc_channel_respond(c->chan,"eno such console",command_id);
        break;
      }
      a=g_new0(edittermContainer,1);
      a->t=t;
      a->c=c;
      a->id=command_id;
      a->spec=conc_termspec_to(&line[5]);
      if (!strcmp(a->spec->server,t->spec.server)) {
        s=g_strdup_printf("e%02x%s",t->id,&line[5]);
        conc_channel_command(t->ts->chan,s,editterm1,a);
        g_free(s);
        break;
      }
      /* FIXME: server had changed, we must close old terminal and
         reopen on new server */
      break;
  }
}

static void
configuration_info(Client *c,gchar *line, gint command_id)
{
  GList *l;
  Term *t;
  gchar *s, *spec;

  switch (line[2]) {
    case 'c': /* console info  - requires response*/
      for (l=terms;l;l=g_list_next(l)) {
        t=(Term *)l->data;
        if (!strcmp(&line[3],t->spec.name)) break;
      }
      if (!l) {
        conc_channel_respond(c->chan,"eno such console",command_id);
        break;
      }
      spec=conc_termspec_from(&t->spec);
      s=g_strdup_printf("d%s",spec); g_free(spec);
      conc_channel_respond(c->chan,s,command_id);
      g_free(s);
      break;
  }
}

static void
configuration_delete(Client *c,gchar *line, gint command_id)
{
  Ts *ts;
  GList *l;
  Term *t;
  GString *tmp_s;
  gchar *s;
  gint i;
  configurationremoveconsoleContainer *a;

  switch (line[2]) {
    case 's': /* delete server - requires response*/
      for (l=tss;l;l=g_list_next(l)) {
        ts=(Ts *)l->data;
        if (!strcmp(&line[3],ts->name)) break;
      }
      if (!l) {
        conc_channel_respond(c->chan,"eno such server",command_id);
        break;
      }
      tmp_s=g_string_new("eterminals are still connected to server:\r");
      i=0;
      for (l=terms;l;l=g_list_next(l)) {
        t=(Term *)l->data;
        if (t->ts==ts) {
          tmp_s=g_string_append(tmp_s,t->spec.name);
          tmp_s=g_string_append_c(tmp_s,'\r');
          i=1;
        }
      }
      if (i) {
        conc_channel_respond(c->chan,tmp_s->str,command_id);
        g_string_free(tmp_s,TRUE);
        break;
      }
      g_string_free(tmp_s,TRUE);
      if (ts->active) {
        ts->active=FALSE;
        conc_channel_close(ts->chan);
      } else {
        g_source_remove(ts->timeout_id);
        s=g_strdup_printf("crs%s",ts->name);
        tss=g_list_remove(tss,ts);
        g_free(ts->name);
        g_free(ts);
        configuration_change(s);
        g_free(s);
        save_config();
      }
      conc_channel_respond(c->chan,"ok",command_id);
      break;
    case 'c': /*remove console - requires response*/
      for (l=terms;l;l=g_list_next(l)) {
        t=(Term *)l->data;
        if (!strcmp(&line[3],t->spec.name)) break;
      }
      if (!l) {
        conc_channel_respond(c->chan,"eno such console",command_id);
        break;
      }
      a=g_new0(configurationremoveconsoleContainer,1);
      t->active=FALSE;
      s=g_strdup_printf("c%02x",t->id);
      a->t=t; a->c=c; a->command_id=command_id;
      conc_channel_command(t->ts->chan,s,configuration_remove_console,a);
      g_free(s);
      break;
  }
}

static void
configuration_new(Client *c,gchar *line, gint command_id)
{
  Ts *ts,*tsx;
  Term *t;
  GList *l;
  GString *tmp_s;
  TermSpec *spec;
  gchar **v;
  gchar *s;

  switch (line[2]) {
    case 's': /* new server */
      ts=g_new0(Ts,1);
      ts->creator=(gpointer)c;
      ts->is_new=command_id;
      ts->local=line[3]=='l'?TRUE:FALSE;
      if (ts->local) {
        ts->name=g_strdup("localhost");
      } else {
        v=g_strsplit(&line[4],"\r",0);
        ts->name=g_strdup(v[0]);
        ts->termserv=g_strdup(v[1]);
        g_strfreev(v);
      }
      for (l=tss;l;l=g_list_next(l)) {
        tsx=(Ts *)l->data;
        if (!strcmp(ts->name,tsx->name)) break;
      }
      if (l) {
        g_free(ts->name);
        if (!ts->local) g_free(ts->termserv);
        s=g_strdup_printf("ethere is already a server named %s",tsx->name);
        conc_channel_respond(c->chan,s,command_id);
        g_free(s);
        break;
      }
      tss=g_list_append(tss,ts);
      ts->timeout_id=g_timeout_add(500,activatets,ts);
      break;
    case 'c': /* new console - requires response */
      spec=conc_termspec_to(&line[3]);
      for (l=terms;l;l=g_list_next(l)) {
        t=(Term *)l->data;
        if (!strcmp(t->spec.name,spec->name)) break;
      }
      if (l) {
        g_free(spec->name);
        g_free(spec->server);
        g_free(spec->device);
        g_free(spec);
        s=g_strdup_printf("ethere is already a console named %s",t->spec.name);
        conc_channel_respond(c->chan,s,command_id);
        g_free(s);
        break;
      }
      t=g_new0(Term,1);
      t->spec=*spec; g_free(spec);
      t->id=idcount++;
      t->is_new=command_id;
      t->creator=c;
      for (l=tss;l;l=g_list_next(l)) {
        ts=(Ts *)l->data;
        if (!strcmp(ts->name,t->spec.server)) break;
      }
      if (l) {
        t->ts=ts;
        terms=g_list_append(terms,t);
        t->timeout_id=g_timeout_add(500,activateterm,t);
      } else {
        tmp_s=g_string_new("");
        g_string_sprintf(tmp_s,"no server %s, disabling console for %s",t->spec.server,t->spec.name);
        g_warning(tmp_s->str);
        g_free(t->spec.name);
        g_free(t->spec.server);
        g_free(t->spec.device);
        g_free(t);
        tmp_s=g_string_prepend_c(tmp_s,'e');
        conc_channel_respond(c->chan,tmp_s->str,command_id);
        g_string_free(tmp_s,TRUE);
      }
      break;
  }
}

static void
configuration(Client *c,gchar *line, gint command_id)
{
  GList *l;
  Ts *ts;
  Term *t;
  gchar *p;
  GString *s;
  gboolean got_one;

  switch (line[1]) {
    case 'R': /*request configuration window data - requires response*/
      s=g_string_new("d");
      got_one=FALSE;
      for (l=tss;l;l=g_list_next(l)) {
        ts=(Ts *)l->data;
        s=g_string_append(s,ts->name);
        s=g_string_append_c(s,'/');
        got_one=TRUE;
      }
      if (got_one) s->str[s->len-1]=':'; else s=g_string_append_c(s,':');
      got_one=FALSE;
      for (l=terms;l;l=g_list_next(l)) {
        t=(Term *)l->data;
        p=g_strdup_printf("%02x%s/",t->id,t->spec.name);
        s=g_string_append(s,p);
        g_free(p);
        got_one=TRUE;
      }
      if (got_one) s=g_string_truncate(s,s->len-1);
      conc_channel_respond(c->chan,s->str,command_id);
      g_string_free(s,TRUE);
      break;
    case 'n': /* new */
      configuration_new(c,line,command_id);
      break;
    case 'd': /* delete */
      configuration_delete(c,line,command_id);
      break;
    case 'i': /* info */
      configuration_info(c,line,command_id);
      break;
    case 'e': /* edit */
      configuration_edit(c,line,command_id);
      break;
  }
}

static gboolean
log_reader(gpointer data)
{
  logreaderContainer *a=(logreaderContainer *)data;
  gchar buf[256];
  gint i;
  GString *line;
  gchar *s;
  
  i=read(a->fd,buf,255);
  if (i<0 && errno==EAGAIN) return TRUE;
  if (!i) {
    i=read(a->errfd,&buf[1],254);
    if (i<1) {
      conc_channel_respond(a->c->chan,"ok",a->command_id);
    } else  {
      buf[i+1]='\0';
      buf[0]='e';
      conc_channel_respond(a->c->chan,buf,a->command_id);
      i=0;
    }
  }
  if (i<0) {
    conc_channel_respond(a->c->chan,"eerror reading log file",a->command_id);
  }
  if (i<=0) {
    close(a->errfd);
    close(a->fd);
    g_string_free(a->s,TRUE);
    g_free(a);
    conc_wait();
    return FALSE;
  }
  buf[i]='\0';
  a->s=g_string_append(a->s,buf);
  while (strstr(a->s->str,"\n")) {
    line=g_string_new("");
    while (a->s->str[0]!='\n') {
      line=g_string_append_c(line,a->s->str[0]);
      a->s=g_string_erase(a->s,0,1);
    }
    a->s=g_string_erase(a->s,0,1); /* drop the \n */
    s=g_strdup_printf("L%02x%s",a->id,line->str);
    g_string_free(line,TRUE);
    conc_channel_write(a->c->chan,s);
    g_free(s);
  }
  return TRUE;
}

static void
list_console_users(Client *cl,gint id,gint command_id)
{
  GList *l;
  Client *c;
  GString *s;

  s=g_string_new("d");
  for (l=clients;l;l=g_list_next(l)) {
    c=(Client *)l->data;
    if (g_list_find(c->terms,GINT_TO_POINTER(id))) {
      s=g_string_append(s,c->name);
      s=g_string_append_c(s,'\r');
    }
  }
  conc_channel_respond(cl->chan,s->str,command_id);
  g_string_free(s,TRUE);
}

static void
read_log(Client *c,gint id,gint lines,gint command_id)
{
  GList *l;
  Term *t;
  gint fromfd[2], errfd[2];
  gchar *fn,*s;
  logreaderContainer *a;

  if (!lines) {
    conc_channel_respond(c->chan,"eno point in sending zero lines",command_id);
    return;
  }
  for (l=terms;l;l=g_list_next(l)) {
    t=(Term *)l->data;
    if (id==t->id) break;
  }
  if (!l) {
    conc_channel_respond(c->chan,"eno such console",command_id);
    g_warning("client attempted read log of non existant term id %02x",id);
    return;
  }
  if (!t->spec.logfile || !strlen(t->spec.logfile)) {
    s=g_strdup_printf("eno log is stored for %s",t->spec.name);
    conc_channel_respond(c->chan,s,command_id);
    g_free(s);
    return;
  }
  if (t->spec.logfile[0]=='/') {
    fn=g_strdup(t->spec.logfile);
  } else {
    fn=g_strdup_printf("%s%s",LOG_DIR,t->spec.logfile);
  }
  s=g_strdup_printf("-%d",lines);
  pipe(fromfd); pipe(errfd);
  switch (fork()) {
    case -1:
      g_warning("could not fork!");
      conc_channel_respond(c->chan,"efailed to fork",command_id);
      return;
    case 0:
      close(0); close(1); close(2);
      close(fromfd[0]); close(errfd[0]);
      dup2(fromfd[1],1); dup2(errfd[1],2);
      execlp("tail","tail",s,fn,NULL);
  }
  g_free(s); g_free(fn);
  close(fromfd[1]); close(errfd[1]);
  fcntl(fromfd[0],F_SETFL,O_NONBLOCK);
  fcntl(errfd[0],F_SETFL,O_NONBLOCK);
  a=g_new0(logreaderContainer,1);
  a->c=c;
  a->s=g_string_new("");
  a->id=id;
  a->fd=fromfd[0];
  a->errfd=errfd[0];
  a->command_id=command_id;
  g_idle_add(log_reader,a);
}

static void
readclient(ConcChannel *chan, gchar *line, gint command_id, gpointer data)
{
  Client *c=(Client *)data;

  switch (line[0]) {
    case 'c': /*configuration*/
      configuration(c,line,command_id);
      break;
    case 'r': /* register to receive a service */
      registerterm(c,line[1],getid(&line[2]));
      break;
    case 'u': /* deregister to stop receiving a service */
      deregisterterm(c,line[1],getid(&line[2]));
      break;
    case 'l': /* list terminals */
      listterms(c,command_id);
      break;
    case 'L': /* request log */
      read_log(c,getid(&line[1]),atoi(&line[3]),command_id);
      break;
    case 'U': /*list users connected to console */
      list_console_users(c,getid(&line[1]),command_id);
    case 'z': /*recieved null*/
      sendterm(getid(&line[1]),0);
      break;
    case 't': /*recieved char*/
      sendterm(getid(&line[1]),line[3]);
      break;
    case 'T': /*received \n*/
      sendterm(getid(&line[1]),'\n');
      break;
    case 'b': /*recieved break*/
      sendbreak(getid(&line[1]));
      break;
    case 'R': /* reset a console */
      resetterm(c,getid(&line[1]),command_id);
      break;
  }
  g_free(line);
}

static void
client_name(ConcChannel *chan,gchar *line,gpointer cd, gpointer d)
{
  Client *c=(Client *)cd;

  if (c->name) free(c->name);
  c->name=line;
  g_message("client id's as %s",c->name);
}

static gboolean
new_client(GIOChannel *chan, GIOCondition cond, gpointer data)
{
  gint ear;
  gint fd;
  gint i;
  struct sockaddr_in isa;
  Client *c;

  ear=g_io_channel_unix_get_fd(chan);
  i=sizeof(isa);
  fd=accept(ear,(struct sockaddr *)&isa,&i);
  if (fd==-1) {
    g_warning("cannot accept connection (%d)",errno);
    return TRUE;
  }
  c=g_new0(Client,1);
  c->chan=conc_channel_new(fd,fd,closeclient,readclient,c,CONC_PASSWORD);
  if (!c->chan) {
    g_free(c);
    return TRUE;
  }
  conc_channel_command(c->chan,"I",client_name,c);
  clients=g_list_append(clients,c);
  return TRUE;
}

static void
start_service(void)
{
  gint ear;
  GIOChannel *chan;
  struct sockaddr_in sa;
  
  ear=socket(AF_INET,SOCK_STREAM,0);
  if (ear==-1) g_error("cannot create socket (%d)",errno);
  bzero(&sa,sizeof sa);
  sa.sin_family=AF_INET;
  sa.sin_port=htons(PORT);
  if (bind(ear,(struct sockaddr *)&sa,sizeof sa)==-1) g_error("cannot bind socket (%d)",errno);
  if (listen(ear,LISTEN_QUEUE)==-1) g_error("cannot listen on socket (%d)",errno);
  chan=g_io_channel_unix_new(ear);
  g_io_add_watch(chan,G_IO_IN,new_client,NULL);
}

static void
check_config(void)
{
  GError *e=NULL;

  gce=gconf_engine_get_default();
  if (gconf_engine_dir_exists(gce,"/apps/conc",&e) && !e) return;
  if (e) {
    g_error("gconf failed: %s",e->message);
    g_error_free(e);
    e=NULL;
    return;
  }
}

static void
conc_log(const gchar *domain, GLogLevelFlags level, const gchar *m, gpointer d)
{
  GIOChannel *chan=(GIOChannel *)d;
  GString *s;
  gint i;

  s=g_string_new(debug?"":get_time(NULL));
  if (level&G_LOG_FLAG_FATAL) s=g_string_append(s," FATAL");
  switch (level&G_LOG_LEVEL_MASK) {
    case G_LOG_LEVEL_ERROR:
      s=g_string_append(s," ERROR: "); break;
    case G_LOG_LEVEL_WARNING:
      s=g_string_append(s," WARNING: "); break;
    case G_LOG_LEVEL_DEBUG:
      s=g_string_append(s," DEBUG: "); break;
    default:
      s=g_string_append_c(s,' '); break;
  }
  s=g_string_append(s,m);
  s=g_string_append_c(s,'\n');
  if (debug) {
    g_print("%s",s->str);
    s=g_string_prepend(s,get_time(NULL));
  }
  g_io_channel_write(chan,s->str,s->len,&i);
  g_string_free(s,TRUE);
}

int
main(int argc, char *argv[])
{
  GError *e=NULL;
  gint fd,i;
  gchar *fn;
  GIOChannel *err;
  struct poptOption options[] = {
    { "usage", '\0', POPT_ARG_NONE, NULL, 1, "Display brief usage message", NULL }, 
    { "help", '?', POPT_ARG_NONE, NULL, 2, "Show this help message", NULL },
    { "debug", 'd', POPT_ARG_NONE, NULL, 3, "Run in debugging mode", NULL },
    { NULL, '\0', 0, NULL, 0 }
  };
  poptContext con;

  gml=g_main_new(FALSE);
  if (!gconf_init(argc,argv,&e)) {
    g_error("failed to init gconf: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  con=poptGetContext("console",argc,(const gchar **)argv,options,0);
  poptSetOtherOptionHelp(con,USAGE);
  while ((i=poptGetNextOpt(con))>0) {
    switch (i) {
      case 1: /*usage*/
        poptPrintUsage(con,stdout,0);
        exit(0);
      case 2: /*help*/
        poptPrintHelp(con,stdout,0);
        exit(0);
      case 3:
        g_print("Entering debug mode...\n");
        debug=TRUE;
        break;
    }
  } 
  if (i<-1) {
    fprintf(stderr,"%s: %s\n", poptBadOption(con,POPT_BADOPTION_NOALIAS),
            poptStrerror(i));
    poptPrintUsage(con,stdout,0);
    fprintf(stderr,HELP,poptGetInvocationName(con));
    exit(1);
  }
  if (poptPeekArg(con)) {
    poptPrintUsage(con,stdout,0);
    fprintf(stderr,HELP,poptGetInvocationName(con));
    exit(1);
  }
  fn=g_strdup_printf("%s%s",LOG_DIR,"conc.log");
  fd=open(fn,O_WRONLY|O_CREAT|O_APPEND,S_IRWXU);
  if (fd<0) g_error("cannot open log file %s",fn);
  g_free(fn);
  if (!debug) { /* don't background in debugging mode */
    switch (fork()) {
      case -1:
        g_error("fork failed (%d)\n",errno);
      case 0:
        close(0); close(1); close(2);
        setsid();
        break;
      default:
        exit(0);
    }
  }
  err=g_io_channel_unix_new(fd);
  g_log_set_handler(G_LOG_DOMAIN,G_LOG_FLAG_RECURSION|G_LOG_FLAG_FATAL|
                                 G_LOG_LEVEL_ERROR|G_LOG_LEVEL_CRITICAL|
                                 G_LOG_LEVEL_WARNING|G_LOG_LEVEL_MESSAGE|
                                 G_LOG_LEVEL_INFO|G_LOG_LEVEL_DEBUG,
                    conc_log,err);
  check_config();
  load_config();
  start_service();
  g_message("server up");
  g_main_run(gml);
  g_message("server down");
  exit(0);
}
