/*
 This file is part of conc.

    Conc is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Conc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with conc; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdarg.h>
#include <gnome.h>

static void
conc_tuple_free(GtkObject *widget, gpointer data)
{
  g_ptr_array_free((GPtrArray *)data,FALSE);
}

/*
  conc_tuple

  w - a widget that the tuple is bound to. then the widget is destroyed the
      tuple will be freed
  ... - a null terminated list of pointers that make up the tuple

  returns: the newly created tuple
*/
GPtrArray *
conc_tuple(GtkWidget *w,...)
{
  GPtrArray *t;
  gpointer p;
  va_list ap;

  t=g_ptr_array_new();
  va_start(ap,w);
  for (p=va_arg(ap,gpointer);p;p=va_arg(ap,gpointer)) {
    g_ptr_array_add(t,p);
  }
  va_end(ap);
  gtk_signal_connect(GTK_OBJECT(w), "destroy",
                     GTK_SIGNAL_FUNC(conc_tuple_free),
                     t);
  return t;
}

/*
  conc_tuple_get

  t - the tuple
  i - the index of the tuple to return (counting from 0)

  returns: the member pointer or NULL if i is out of bounds
*/
gpointer
conc_tuple_get(GPtrArray *t,guint i)
{
  if (i>=t->len) g_warning("tried to access nonexistant tuple member");
  return g_ptr_array_index(t,i);
}
